/*
!C******************************************************************
!Description:
	Functions to handle input from HDF5 files.

!Revision History:
!Team-unique Header:
!References and Credits:

	W. Schroeder, University of Maryland
	W. Walsh, University of Maryland

!Design Notes:

!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>


#include "hdf5.h"

#include "const.h"
#include "types.h"

#include "input.h"
#include "error.h"
#include "misc.h"


#define OPTIONAL_SCALE_OFFSET 1



static int init_input_sds(hid_t sd_id, char *sds_name, hid_t *sds_index_ptr);
static void unscale(uint16 **ScaledVal, float **UnscaledVal, double scale, double offset, int scan_lines);
extern int verbose;


/* Open all SDSs and store the indexs as well as scale/offsets in the InputInfo struct */
int InitInputHDF(hid_t sd_id_i01, hid_t sd_id_i02, hid_t sd_id_i03, hid_t sd_id_i04, hid_t sd_id_i05, hid_t sd_id_gitco, hid_t sd_id_m13_agg, hid_t sd_id_m13_unagg, InputInfo *in_info){
	int st;
	float *buff;
	in_info->sd_id_i01 = sd_id_i01;
	in_info->sd_id_i02 = sd_id_i02;
	in_info->sd_id_i03 = sd_id_i03;
	in_info->sd_id_i04 = sd_id_i04;
	in_info->sd_id_i05 = sd_id_i05;
	in_info->sd_id_gitco = sd_id_gitco;
	in_info->sd_id_m13_agg = sd_id_m13_agg;
	in_info->sd_id_m13_unagg = sd_id_m13_unagg;


	/* ---------- I-bands ---------- */

	/* I1 reflectance */
	if ( (st = init_input_sds(sd_id_i01, "/All_Data/VIIRS-I1-SDR_All/Reflectance",
		&in_info->sds_index_ref1)) ) return error(st);

	hid_t dataspace_id = H5Dget_space(in_info->sds_index_ref1);
	hssize_t elements = H5Sget_simple_extent_npoints(dataspace_id);

	/* Number of image lines should be a multiple of 32 to match the 32
	detectors acquiring data during each scan. */
	if ((elements/SAMPLES_375M) % 32) return error(ERR_IMAGE_LINES);

	/* 32 lines per full VIIRS scan */
	in_info->nscans = (elements/SAMPLES_375M) / 32;
	printf("Scans: %d\n", in_info->nscans);

	/* I1 reflectance factors */
	if ( (st = init_input_sds(sd_id_i01, "/All_Data/VIIRS-I1-SDR_All/ReflectanceFactors",
		&in_info->sds_index_ref1_fac)) ) return error(st);

	/* Malloc space for scale and offset factors */
	dataspace_id = H5Dget_space(in_info->sds_index_ref1_fac);
	elements = H5Sget_simple_extent_npoints(dataspace_id);
	buff = malloc(sizeof(float)*elements);

	/* Get Scale and Offset */
	st = H5Dread(in_info->sds_index_ref1_fac, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0]);
	in_info->ScaleR1 = buff[0];
	in_info->OffsetR1 = buff[1];
	//printf("%f - %f \n", in_info->ScaleR1, in_info->OffsetR1);
 
	/* I1 QA */ 
	if ( (st = init_input_sds(sd_id_i01, "/All_Data/VIIRS-I1-SDR_All/QF1_VIIRSIBANDSDR",
		&in_info->sds_index_QF1)) ) return error(st);


	// I2 reflectance 
	if ( (st = init_input_sds(sd_id_i02, "/All_Data/VIIRS-I2-SDR_All/Reflectance",
		&in_info->sds_index_ref2)) ) return error(st);

	/* I2 reflectance factors */
	if ( (st = init_input_sds(sd_id_i02, "/All_Data/VIIRS-I2-SDR_All/ReflectanceFactors",
		&in_info->sds_index_ref2_fac)) ) return error(st);

	/* Get Scale and Offset */
	st = H5Dread(in_info->sds_index_ref2_fac, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0]);
	in_info->ScaleR2 = buff[0];
	in_info->OffsetR2 = buff[1];
	//printf("%f - %f \n", in_info->ScaleR2, in_info->OffsetR2);

	// I2 QA
	if ( (st = init_input_sds(sd_id_i02, "/All_Data/VIIRS-I2-SDR_All/QF1_VIIRSIBANDSDR",
		&in_info->sds_index_QF2)) ) return error(st);

	// I3 reflectance
	if ( (st = init_input_sds(sd_id_i03, "/All_Data/VIIRS-I3-SDR_All/Reflectance",
		&in_info->sds_index_ref3)) ) return error(st);

	/* I3 reflectance factors */
	if ( (st = init_input_sds(sd_id_i03, "/All_Data/VIIRS-I3-SDR_All/ReflectanceFactors",
		&in_info->sds_index_ref3_fac)) ) return error(st);

	/* Get Scale and Offset */
	st = H5Dread(in_info->sds_index_ref3_fac, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0]);
	in_info->ScaleR3 = buff[0];
	in_info->OffsetR3 = buff[1];
	//printf("%f - %f \n", in_info->ScaleR3, in_info->OffsetR3);

	// I3 QA 
	if ( (st = init_input_sds(sd_id_i03, "/All_Data/VIIRS-I3-SDR_All/QF1_VIIRSIBANDSDR",
		&in_info->sds_index_QF3)) ) return error(st);

	// I4 brightness tempature 
	if ( (st = init_input_sds(sd_id_i04, "/All_Data/VIIRS-I4-SDR_All/BrightnessTemperature",
		&in_info->sds_index_bt4)) ) return error(st);

	/* I4 brightness tempature factors */
	if ( (st = init_input_sds(sd_id_i04, "/All_Data/VIIRS-I4-SDR_All/BrightnessTemperatureFactors",
		&in_info->sds_index_bt4_fac)) ) return error(st);

	/* Get Scale and Offset */
	st = H5Dread(in_info->sds_index_bt4_fac, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0]);
	in_info->ScaleT4 = buff[0];
	in_info->OffsetT4 = buff[1];
	//printf("%f - %f \n", in_info->ScaleT4, in_info->OffsetT4);

	// I4 QA 
	if ( (st = init_input_sds(sd_id_i04, "/All_Data/VIIRS-I4-SDR_All/QF1_VIIRSIBANDSDR",
		&in_info->sds_index_QF4)) ) return error(st);

	// I5 brightness tempature 
	if ( (st = init_input_sds(sd_id_i05, "/All_Data/VIIRS-I5-SDR_All/BrightnessTemperature",
		&in_info->sds_index_bt5)) ) return error(st);

	/* I5 brightness tempature factors */
	if ( (st = init_input_sds(sd_id_i05, "/All_Data/VIIRS-I5-SDR_All/BrightnessTemperatureFactors",
		&in_info->sds_index_bt5_fac)) ) return error(st);

	/* Get Scale and Offset */
	st = H5Dread(in_info->sds_index_bt5_fac, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0]);
	in_info->ScaleT5 = buff[0];
	in_info->OffsetT5 = buff[1];
	//printf("%f - %f \n", in_info->ScaleT5, in_info->OffsetT5);

	// I5 QA
	if ( (st = init_input_sds(sd_id_i05, "/All_Data/VIIRS-I5-SDR_All/QF1_VIIRSIBANDSDR",
		&in_info->sds_index_QF5)) ) return error(st);
	
	// ---------- M-band ---------- 
	if(sd_id_m13_agg != NULL){
		// M13 Agg radiance
		if ( (st = init_input_sds(sd_id_m13_agg, "/All_Data/VIIRS-M13-SDR_All/Radiance",
			&in_info->sds_index_rad13_agg)) ) return error(st);

		// M13 QA
		if ( (st = init_input_sds(sd_id_m13_agg, "/All_Data/VIIRS-M13-SDR_All/QF1_VIIRSMBANDSDR",
			&in_info->sds_index_QF13)) ) return error(st);
	}else{
		in_info->sds_index_rad13_agg = NULL;
		in_info->sds_index_QF13 = NULL;
	}

	if(sd_id_m13_unagg != NULL){
		// M13 Unagg brightness tempature 
		if ( (st = init_input_sds(sd_id_m13_unagg, "/All_Data/VIIRS-DualGain-Cal-IP_All/radiance_6",
			&in_info->sds_index_rad13_unagg)) ) return error(st);
	}else{
		in_info->sds_index_rad13_unagg = NULL;
	}

	/* BeginningDate */
	hid_t atype = H5Tcopy(H5T_C_S1);
	H5Tset_size(atype, 9);
   	hid_t dataset = H5Dopen(sd_id_gitco, "/Data_Products/VIIRS-IMG-GEO-TC/VIIRS-IMG-GEO-TC_Aggr", H5P_DEFAULT);
   	hid_t attr = H5Aopen_name(dataset, "AggregateBeginningDate");
   	st  = H5Aread(attr, atype, in_info->BeginningDate);
   	st =  H5Aclose(attr);

	H5Tset_size(atype, 15);
   	dataset = H5Dopen(sd_id_gitco, "/Data_Products/VIIRS-IMG-GEO-TC/VIIRS-IMG-GEO-TC_Aggr", H5P_DEFAULT);
   	attr = H5Aopen_name(dataset, "AggregateBeginningTime");
   	st  = H5Aread(attr, atype, in_info->BeginningTime);
   	st =  H5Aclose(attr);


	// ---------- Geolocation ---------- 
	if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/Latitude",
		&in_info->sds_index_lat)) ) return error(st);

	if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/Longitude",
		&in_info->sds_index_lon)) ) return error(st);

	if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/SolarZenithAngle",
		&in_info->sds_index_solzen)) ) return error(st);

	if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/SolarAzimuthAngle",
		&in_info->sds_index_solaz)) ) return error(st);

	if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/SatelliteZenithAngle",
		&in_info->sds_index_viewzen)) ) return error(st);

	if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/SatelliteAzimuthAngle",
		&in_info->sds_index_viewaz)) ) return error(st);

	if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/MidTime",
		&in_info->sds_index_midtime)) ) return error(st);

        if ( (st = init_input_sds(sd_id_gitco, "/All_Data/VIIRS-IMG-GEO-TC_All/QF2_VIIRSSDRGEO",
                &in_info->sds_index_QFGeo)) ) return error(st);


	free(buff);

	return 0;
}


/* Gets sds index */
static int init_input_sds(hid_t sd_id, char *sds_name, hid_t *sds_index_ptr)
{
	hid_t sds_index;

	sds_index = H5Dopen2(sd_id, sds_name, H5P_DEFAULT);
	if (sds_index <= 0){
		printf("Error: init_input_sds\n");
		return -1;
	}	

	*sds_index_ptr = sds_index;
	return 0;
}





/* Reads all the input SDSs and if needed unscales them */
/* (Need checks for all mallocs) */
int ReadInputs(InputInfo *in_info, Juggernaut *jugg){	

	int i, j, st;

	/* buff is used to get scaled reflectance data */
	uint16 **buff;
	if ( !(buff = (uint16**)malloc(jugg->number_of_scan_lines*sizeof(uint16*))) )
		return error(ERR_F_MEM);
	if ( !(buff[0] = (uint16*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(uint16))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		buff[i] = buff[i-1] + SAMPLES_375M;
	}

	/* ----- I-band ----- */
	/* I1 reflectance */
	if ( !(jugg->R1 = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->R1[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->R1[i] = jugg->R1[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_ref1, H5T_NATIVE_USHORT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0][0]);
	unscale(buff, jugg->R1, in_info->ScaleR1, in_info->OffsetR1, jugg->number_of_scan_lines);

	/* I1 QA */
	if ( !(jugg->QF1 = (uint8**)malloc(jugg->number_of_scan_lines*sizeof(uint8*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->QF1[0] = (uint8*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(uint8))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->QF1[i] = jugg->QF1[i-1] + SAMPLES_375M;
	}
	
	st = H5Dread(in_info->sds_index_QF1, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->QF1[0][0]);	

	/* I2 reflectance */
	if ( !(jugg->R2 = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->R2[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->R2[i] = jugg->R2[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_ref2, H5T_NATIVE_USHORT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0][0]);
	unscale(buff, jugg->R2, in_info->ScaleR2, in_info->OffsetR2, jugg->number_of_scan_lines);

	/* I2 QA */
	if ( !(jugg->QF2 = (uint8**)malloc(jugg->number_of_scan_lines*sizeof(uint8*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->QF2[0] = (uint8*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(uint8))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->QF2[i] = jugg->QF2[i-1] + SAMPLES_375M;
	}
	
	st = H5Dread(in_info->sds_index_QF2, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->QF2[0][0]);
	
	/* I3 reflectance */
	if ( !(jugg->R3 = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->R3[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->R3[i] = jugg->R3[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_ref3, H5T_NATIVE_USHORT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0][0]);
	unscale(buff, jugg->R3, in_info->ScaleR3, in_info->OffsetR3, jugg->number_of_scan_lines);

	/* I3 QA */
	if ( !(jugg->QF3 = (uint8**)malloc(jugg->number_of_scan_lines*sizeof(uint8*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->QF3[0] = (uint8*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(uint8))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->QF3[i] = jugg->QF3[i-1] + SAMPLES_375M;
	}
	
	st = H5Dread(in_info->sds_index_QF3, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->QF3[0][0]);

	/* I4 brightness tempature */
	if ( !(jugg->T4 = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->T4[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->T4[i] = jugg->T4[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_bt4, H5T_NATIVE_USHORT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0][0]);
	unscale(buff, jugg->T4, in_info->ScaleT4, in_info->OffsetT4, jugg->number_of_scan_lines);

	/* I4 QA */
	if ( !(jugg->QF4 = (uint8**)malloc(jugg->number_of_scan_lines*sizeof(uint8*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->QF4[0] = (uint8*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(uint8))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->QF4[i] = jugg->QF4[i-1] + SAMPLES_375M;
	}
	
	st = H5Dread(in_info->sds_index_QF4, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->QF4[0][0]);

	/* I5 brightness tempature */
	if ( !(jugg->T5 = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->T5[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->T5[i] = jugg->T5[i-1] + SAMPLES_375M;
	}


	st = H5Dread(in_info->sds_index_bt5, H5T_NATIVE_USHORT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &buff[0][0]);
	unscale(buff, jugg->T5, in_info->ScaleT5, in_info->OffsetT5, jugg->number_of_scan_lines);
	
	/* I5 QA */
	if ( !(jugg->QF5 = (uint8**)malloc(jugg->number_of_scan_lines*sizeof(uint8*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->QF5[0] = (uint8*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(uint8))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->QF5[i] = jugg->QF5[i-1] + SAMPLES_375M;
	}
	
	st = H5Dread(in_info->sds_index_QF5, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->QF5[0][0]);

	/* ----- I-band Geo ----- */
	/* latitude */
	if ( !(jugg->lat = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->lat[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->lat[i] = jugg->lat[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_lat, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->lat[0][0]);
	
	/* longitude */
	if ( !(jugg->lon = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->lon[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->lon[i] = jugg->lon[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_lon, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->lon[0][0]);

	/* view zenith angle */
	if ( !(jugg->ViewZenAng = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->ViewZenAng[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->ViewZenAng[i] = jugg->ViewZenAng[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_viewzen, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->ViewZenAng[0][0]);

	/* solar zenith angle */
	if ( !(jugg->SolZenAng = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->SolZenAng[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->SolZenAng[i] = jugg->SolZenAng[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_solzen, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->SolZenAng[0][0]);
	
	/* view azimuth angle */
	if ( !(jugg->ViewAzAng = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->ViewAzAng[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->ViewAzAng[i] = jugg->ViewAzAng[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_viewaz, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->ViewAzAng[0][0]);

	/* solar azimuth angle */
	if ( !(jugg->SolAzAng = (float**)malloc(jugg->number_of_scan_lines*sizeof(float*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->SolAzAng[0] = (float*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(float))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->SolAzAng[i] = jugg->SolAzAng[i-1] + SAMPLES_375M;
	}

	st = H5Dread(in_info->sds_index_solaz, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->SolAzAng[0][0]);

	/* mid time */
	if ( !(jugg->MidTime = (long*)malloc(in_info->nscans*sizeof(long))) )
		return error(ERR_F_MEM);
	st = H5Dread(in_info->sds_index_midtime, H5T_NATIVE_LONG, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->MidTime[0]);

	/* qa */
	if ( !(jugg->QFGeo = (uint8**)malloc(jugg->number_of_scan_lines*sizeof(uint8*))) )
		return error(ERR_F_MEM);
	if ( !(jugg->QFGeo[0] = (uint8*)malloc(jugg->number_of_scan_lines*SAMPLES_375M*sizeof(uint8))) )
		return error(ERR_F_MEM);
	for (i=1;i<jugg->number_of_scan_lines;i++){
		jugg->QFGeo[i] = jugg->QFGeo[i-1] + SAMPLES_375M;
	}
	
	st = H5Dread(in_info->sds_index_QFGeo, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, &jugg->QFGeo[0][0]);

	return 0;
}


/* Reads SDS into buff */
int read_band(hid_t sds_index, int type, void *buff){
	int st;
	st = H5Dread(sds_index, type, H5S_ALL, H5S_ALL, H5P_DEFAULT, buff);

	return 0;
}



/* Unscale scaled brightness temperatures and reflectances into physical values.
   Missing data will be flagged with a value of -1. */
static void unscale(uint16 **ScaledVal, float **UnscaledVal, double scale, double offset, int scan_lines){
	int i, j;
	const float missval = 0;
	//printf("scale: %f, offset: %f\n", scale, offset);
	for(i=0;i<scan_lines;i++){
		for(j=0;j<SAMPLES_375M;j++){
			if (ScaledVal[i][j] < 65528){
				UnscaledVal[i][j] = ScaledVal[i][j]*scale + offset;
			}else{
				UnscaledVal[i][j] = missval;
			}		
		}
	}
}

















